@@ -12,6 +12,7 @@ try:
 except ModuleNotFoundError:
     _HAS_CUDAMAT=False
 
+from . import coretools
 from .coretools import ensure_np_array
 
 def _softplus(x):
@@ -101,7 +102,7 @@ class ANN:
          self.cu_b=[]
          for i in range(len(self.b)):
              self.cu_b.append(cm.CUDAMatrix(np.reshape(self.b[i],(1,-1))))
-    
+
     # set array of parameters
     def setpar(self,par):
         assert len(par)==self.npar
@@ -141,7 +142,7 @@ class ANN:
 
         assert x.shape[1]==self.narg
         f,df_dW,df_db=self.deriv(x)
-        der=np.zeros((x.shape[0],self.npar))
+        der=np.zeros((x.shape[0],self.npar),dtype=f.dtype)
         n=0
         for i in range(len(self.W)):
             m=np.prod(self.W[i].shape)
@@ -174,10 +175,111 @@ class ANN:
                 cu_x.add_row_vec(self.cu_b[i])
                 if i+1<len(self.cu_W):
                     self._activation(cu_x)
-            x=np.array(cu_x.asarray(),dtype=np.float)
-         
+            x=np.array(cu_x.asarray())
+
         return x[:,0]
 
+
+    def forward(self,x):
+        x = ensure_np_array(x)
+
+        if len(x.shape)==1:
+            f,df_dW,df_db = self.deriv(x.reshape((1,len(x))))
+            for i in range(len(df_dW)):
+                df_dW[i]=df_dW[i][0]
+                df_db[i]=df_db[i][0]
+            return f[0], df_dW, df_db
+        elif len(x.shape)>2:
+            raise TypeError("Incorrectly shaped x")
+
+        # allocate hidden nodes
+        h=[None]*len(self.layers) # hidden nodes
+        ht=[None]*len(self.layers) # non-linear functions of hidden nodes
+
+        if not self.cuda:
+            # forward propagation
+            ht[0]=x.copy()
+            for i in range(len(self.layers)-1):
+                h[i+1]=np.matmul(ht[i],self.W[i])+self.b[i]
+                ht[i+1] = self._activation(h[i+1])
+            f=(np.matmul(ht[-1],self.W[-1])+self.b[-1])[:,0]
+        else:
+            ht[0]=cm.CUDAMatrix(x)
+            for i in range(len(self.cu_W)-1):
+                h[i+1]=cm.dot(ht[i],self.cu_W[i])
+                h[i+1].add_row_vec(self.cu_b[i])
+                ht[i+1]=h[i+1].copy()
+                self._activation(ht[i+1])
+            f=cm.dot(ht[-1],self.cu_W[-1])
+            f.add_row_vec(self.cu_b[-1])
+
+        class State(coretools.Result):
+            pass
+        return State(f=f,h=h,ht=ht)
+
+    def backward(self,deriv,hidden):
+        # allocate derivatives
+        df_dW=[None]*len(self.layers)
+        df_db=[None]*len(self.layers)
+
+        if not self.cuda:
+            df_db[-1]=np.ones((len(hidden.f),1))
+            df_dW[-1]=np.matmul(deriv,hidden.ht[-1])[:,np.newaxis]
+            for i in reversed(range(len(self.layers)-1)):
+                df_db[i]=np.matmul(df_db[i+1],self.W[i+1].T) * self._dactivation(hidden.h[i+1])
+                df_dW[i]=np.einsum("i,ij,ik->jk",deriv,hidden.ht[i],df_db[i])
+            for i in range(len(self.layers)):
+                df_db[i]=np.matmul(deriv,df_db[i])
+
+        else:
+           if not isinstance(deriv,cm.CUDAMatrix):
+               deriv=deriv.reshape((1,-1))
+               deriv=cm.CUDAMatrix(deriv)
+
+           if deriv.shape[1]==1:
+               deriv=deriv.transpose()
+
+           vec=deriv.shape[1]
+ 
+           df_db[-1]=cm.CUDAMatrix(np.ones((vec,1)))
+           df_dW[-1]=cm.dot(deriv,hidden.ht[-1]).transpose()
+ 
+           for i in reversed(range(len(self.layers)-1)):
+               self._dactivation(hidden.h[i+1])
+               df_db[i]=cm.dot(df_db[i+1],self.cu_W[i+1].transpose())
+               df_db[i].mult(hidden.h[i+1])
+ 
+               hidden.ht[i].mult_by_col(deriv.transpose())
+               df_dW[i]=cm.dot(hidden.ht[i].transpose(),df_db[i])
+ 
+           for i in range(len(self.layers)):
+               df_db[i]=cm.dot(deriv,df_db[i])
+           for i in range(len(df_db)):
+               df_db[i]=df_db[i].asarray()[0,:]
+           for i in range(len(df_db)):
+               df_dW[i]=df_dW[i].asarray()
+
+        return df_dW,df_db
+
+    def backward_par(self,deriv,hidden):
+        df_dW,df_db=self.backward(deriv,hidden)
+        if not self.cuda:
+            der=np.zeros(self.npar,dtype=hidden.f.dtype)
+        else:
+            der=np.zeros(self.npar,dtype=df_dW[0].dtype)
+        n=0
+        for i in range(len(self.W)):
+            m=np.prod(self.W[i].shape)
+            der[n:n+m]=df_dW[i].flatten()
+            n+=m
+        for i in range(len(self.b)):
+            m=np.prod(self.b[i].shape)
+            der[n:n+m]=df_db[i]
+            n+=m
+        assert(n==self.npar)
+        return der
+
+
     # compute derivatives with respect to parameters
     def deriv(self,x):
 
@@ -207,7 +309,7 @@ class ANN:
         if not self.cuda:
 
             # forward propagation
-            ht[0]=+x
+            ht[0]=x.copy()
 
             for i in range(len(self.layers)-1):
                 h[i+1]=np.matmul(ht[i],self.W[i])+self.b[i]
@@ -239,13 +341,10 @@ class ANN:
 
             # backward propagation
 
-            for i in range(len(self.cu_W)):
-                ht[i]=ht[i].asarray()
-
             df_db_host=[None]*len(self.layers)
 
             df_db[-1]=cm.CUDAMatrix(np.ones((vec,1)))
-            df_dW[-1]=ht[-1][:,:,np.newaxis]
+            df_dW[-1]=ht[-1].asarray()[:,:,np.newaxis]
 
             df_db_host[-1]=df_db[-1].asarray()
 
@@ -253,19 +352,14 @@ class ANN:
                 self._dactivation(h[i+1])
                 df_db[i]=cm.dot(df_db[i+1],self.cu_W[i+1].transpose())
                 df_db[i].mult(h[i+1])
-                df_db_host[i]=df_db[i].asarray()
+                df_db_host[i]=df_db[i].asarray() # should be moved to CPU anyway
  
-                # this is still on CPU:
-                df_dW[i]=ht[i][:,:,np.newaxis]*df_db_host[i][:,np.newaxis,:]
+                # this is still on CPU, but could be done on GPU avoiding the movement of ht[i]
+                df_dW[i]=ht[i].asarray()[:,:,np.newaxis]*df_db_host[i][:,np.newaxis,:]
 
             df_db=df_db_host
 
-            for i in range(len(df_dW)):
-                df_dW[i]=np.array(df_dW[i],dtype=np.float)
-            for i in range(len(df_db)):
-                df_db[i]=np.array(df_db[i],dtype=np.float)
-            f=np.array(f,dtype=np.float)
-            
+            f=np.array(f)
 
         return f,df_dW,df_db
 
